import json
import math
from datetime import datetime
from functools import lru_cache
from typing import Optional

import ffmpeg
import pandas as pd
from dirlin import Path, Folder


class Video:
    def __init__(self, file_path: Path):
        """
        An object representing a Video. Holds information regarding the video file.

        Attributes:
            - filepath: the full path, we use this as arguments for ffmpeg functions
            - filename: the name of the video file (i.e. example.mp4)
            - stem: name of the video without the extension (i.e. example)
            - folder_path: the parent folder path (i.e. main_directory/*)
            - meta: holds data regarding the video file

        :param file_path: the pathlib.Path file path to the Video
        """
        self.filename = str(file_path.name)
        self.stem = str(file_path.stem)
        self.folder_path = str(file_path.parent)
        self._filepath = file_path

        # Creation of the Metadata
        _size = f"{round(file_path.stat().st_size * 0.000001, 2)} MB"
        _last_opened = f"{datetime.fromtimestamp(file_path.stat().st_atime)}"
        _last_modified = f"{datetime.fromtimestamp(file_path.stat().st_mtime)}"
        try:
            _created = f"{datetime.fromtimestamp(file_path.stat().st_birthtime)}"
        except AttributeError:
            _created = "Unknown"

        # metadata from ffmpeg.probe
        try:
            _probe = ffmpeg.probe(str(file_path))
            _vs = next((stream for stream in _probe["streams"] if stream['codec_type'] == 'video'), None)
        except Exception as e:
            print(e)
            _vs = dict()
            _vs['duration'] = 60
            _vs['avg_frame_rate'] = '30'
            _vs['width'] = 0
            _vs['height'] = 0

        _duration = float(_vs['duration'])
        _fps = eval(_vs["avg_frame_rate"])
        _total_frames = int(_duration * _fps)
        _resolution = f"{_vs['width']}x{_vs['height']}"

        _meta = {
            "size": _size,
            "last_opened": _last_opened,
            "last_modified": _last_modified,
            "created": _created,
            "duration": _duration / 60,
            "frames": _total_frames,
            "resolution": _resolution
        }
        self.meta = json.dumps(_meta)

    @property
    def filepath(self):
        """
        Full path of the video file. Includes the parent and the filename. Used for ffmpeg functions.
        """
        return str(self._filepath)

    @property
    def size(self):
        """
        Size in mb of the file
        """
        return json.loads(self.meta)["size"]

    @property
    def date_last_opened(self):
        """
        date the file was last opened
        """
        return json.loads(self.meta)["last_opened"]

    @property
    def date_last_modified(self):
        """
        date the file was last modified
        """
        return json.loads(self.meta)["last_modified"]

    @property
    def date_created(self):
        """
        date the file was created
        """
        return json.loads(self.meta)["created"]

    @property
    def duration(self):
        """
        duration of the video
        """
        return json.loads(self.meta)["duration"]

    @property
    def frames(self):
        """
        number of frames
        """
        return json.loads(self.meta)["frames"]

    @property
    def resolution(self):
        """
        resolution of the video

        """
        return json.loads(self.meta)["resolution"]

    def __repr__(self):
        return f"filename: {self.filename}: {self.meta}"


class VideoEditor:
    @classmethod
    def join_and_save(
            cls,
            base_video: Video,
            new_video: Video,
            output_path: str,
            bv_kw: Optional[dict] = None,
            nv_kw: Optional[dict] = None,
            concat_kw: Optional[dict] = None
    ):
        """
        Appends a new video at the end of the base video. Allows for keywords to be added
        so that you can add transitions to the videos.

        :param base_video: The video the function will append to
        :param new_video: the video that is going to be added at the end of the base video
        :param output_path: where to save the video and the filename
        :param bv_kw: base video keyword arguments for the input function
        :param nv_kw: new video keyword arguments for the input function
        :param concat_kw: keyword arguments for the concat function

        :return:
        """
        kw_args = (bv_kw, nv_kw, concat_kw)
        use_kw_args = [dict() if not kw else kw for kw in kw_args]

        in1 = ffmpeg.input(str(base_video.filepath), **use_kw_args[0])
        in2 = ffmpeg.input(str(new_video.filepath), **use_kw_args[1])

        # base video filtering and updates
        v1 = in1.video
        a1 = in1.audio

        # new video filtering and updates
        v2 = in2.video
        a2 = in2.audio

        joined = ffmpeg.concat(v1, a1, v2, a2, v=1, a=1, **use_kw_args[2]).node
        out = ffmpeg.output(joined[0], joined[1], output_path)
        out.run()

    @classmethod
    def cut_and_save(
            cls,
            video: Video,
            output_path: str,
            start: int,
            end: Optional[int] = None,
            duration: Optional[int] = None):
        """
        Arguments are made in seconds, so you might need to convert for now.
        Will fix in the future to be in minutes, hours, etc.

        If both end and duration are given, the function will default to duration.

        ffmpeg.trim(stream, **kwargs)
        Trim the input so that the output contains one continuous subpart of the input.

        Parameters
        start – Specify the time of the start of the kept section
        end – Specify the time of the first frame that will be dropped
        start_pts – This is the same as start, except this option sets the start timestamp in timebase units
        end_pts – This is the same as end, except this option sets the end timestamp in timebase units
        duration – The maximum duration of the output in seconds.
        start_frame – The number of the first frame that should be passed to the output.
        end_frame – The number of the first frame that should be dropped.
        """
        in1 = ffmpeg.input(str(video.filepath))

        if not end and not duration:
            # args not given
            trimmed_vid = in1.video.trim(start=start).setpts('PTS-STARTPTS')
            trimmed_aud = in1.audio.filter_('atrim', start=start).filter_('asetpts', 'PTS-STARTPTS')
        elif all([end, duration]) or duration:
            # duration is given so takes priority
            trimmed_vid = in1.video.trim(start=start, end=start + duration).setpts('PTS-STARTPTS')
            trimmed_aud = (
                in1.audio
                .filter_('atrim', start=start, end=start + duration)
                .filter_('asetpts', 'PTS-STARTPTS')
            )
        else:
            trimmed_vid = in1.video.trim(start=start, end=end).setpts('PTS-STARTPTS')
            trimmed_aud = (
                in1.audio
                .filter_('atrim', start=start, end=end)
                .filter_('asetpts', 'PTS-STARTPTS')
            )

        joined = ffmpeg.concat(trimmed_vid, trimmed_aud, v=1, a=1).node
        out = ffmpeg.output(joined[0], joined[1], output_path)
        out.run()

    @classmethod
    def generate_thumbnail_sheet(
            cls,
            video: Video,
            output_path: str,
            columns: Optional[int] = 5,
            rows: Optional[int] = 4
    ):
        thumbnail_width = 320
        thumbnail_length = 240

        # frames between each snapshot
        frames = max(1, math.floor(video.frames / (columns * rows)))
        metadata_text = f"Filename: {video.filename}, File Size: {video.size}"

        in1 = ffmpeg.input(video.filepath)
        out = (
            in1
            .filter('select', fr'not(mod(n, {frames}))')
            .filter('scale', thumbnail_width, thumbnail_length)
            .filter('drawtext', text='%{pts:hms}', x=10, y=10, fontsize=24, fontcolor='white')
            .filter('tile', f'{columns}x{rows}')
            .output(output_path, vsync='vfr', vframes=1)
        )
        out.run()


class ApplicationConfig:
    @classmethod
    def return_current_directory(cls):
        """
        Function for returning the current working directory

        """
        get_path = str(Path(__file__).parent.parent / "app_config.json")

        with open(get_path, "r") as f:
            config_options = json.loads(f.read())
        if config_options["folder"] == "default":
            path = Path(__file__).parent.parent / "files"
        else:
            path = config_options["folder"]
        return path

    @classmethod
    def change_current_directory(cls, cwd: str | Path) -> bool:
        """
        Function for changing the current working directory and writes in the app_config JSON file
        :param cwd: the working directory you want to update to. Can set to 'default' to change it back
        :return: return True when complete
        """
        get_path = str(Path(__file__).parent.parent / "app_config.json")

        try:
            if isinstance(cwd, str) and cwd != "default":
                cwd = Path(cwd)
            if cwd != "default":
                assert cwd.exists()
        except AssertionError:
            raise AssertionError(f"The directory {cwd} does not exist.")

        with open(get_path, "r+") as f:
            config_options = json.load(f)
            config_options['folder'] = str(cwd)
            f.seek(0)
            json.dump(config_options, f, indent=4)
            f.truncate()
        return True

    @classmethod
    @lru_cache()
    def pull_from_directory(cls, file_ext=".mp4", *index_file_args, **index_file_kw) -> list[Video]:
        """
        Pulls from the app_config.json file in the root of the app, to determine which folder to
        pull the files from. The function itself returns a Folder object. If missing, creates a folder

        :return:
        """
        get_path = str(Path(__file__).parent.parent / "app_config.json")

        with open(get_path, "r") as j_file:
            config_options = json.loads(j_file.read())

        if config_options["folder"] == "default":
            default_path = Path(__file__).parent.parent / "files"
            if not default_path.exists():
                default_path.mkdir(parents=True, exist_ok=True)
            folder_path = Folder(default_path)
        else:
            folder_path = Folder(config_options["folder"])
        videos = [Video(v) for v in folder_path.index_files(file_ext=file_ext, *index_file_args, **index_file_kw)]
        return videos

    @classmethod
    def as_dataframe(cls, *index_file_args, **index_file_kw) -> pd.DataFrame:
        """
        returns the video list as a dataframe. this allows the sorting of the table in the main application.

        :param file_ext: the extension we are searching for in the folders
        :param index_file_args: arguments for the index_files function
        :param index_file_kw: keyword arguments for the index_files function
        :return: dataframe representation of the file and the metadata
        """
        videos = cls.pull_from_directory(*index_file_args, **index_file_kw)

        df_mapping = {
            "idx": [],
            "filename": [],
            "folder_path": [],
            "size": [],
            "date_created": [],
            "date_modified": [],
            "date_accessed": [],
            "duration": [],
            "resolution": []
        }
        for idx, video in enumerate(videos):
            df_mapping["idx"].append(idx)
            df_mapping["filename"].append(video.filename)
            df_mapping["folder_path"].append(video.folder_path)
            df_mapping["size"].append(video.size)
            df_mapping["date_created"].append(video.date_created)
            df_mapping["date_modified"].append(video.date_last_modified)
            df_mapping["date_accessed"].append(video.date_last_opened)
            df_mapping["duration"].append(video.duration)
            df_mapping["resolution"].append(video.resolution)

        df = pd.DataFrame.from_dict(df_mapping)
        return df



import datetime
import random
from typing import Optional

import typer
from rich import box
from rich.align import Align
from rich.prompt import Prompt, Confirm

from rich.console import Console
from rich.table import Table
from rich.text import Text

from stitchi.backend.classes import VideoEditor, ApplicationConfig

console = Console()
app = typer.Typer(no_args_is_help=True)


@app.command(short_help="stiches a video together. takes base index, addon index, and output name")
def add(base_video_index: int, add_video_index: int, output_name: Optional[str] = None):
    videos = ApplicationConfig.pull_from_directory(recurse=True)
    v1 = videos[base_video_index]
    v2 = videos[add_video_index]

    if output_name:
        if output_name[-4:] != ".mp4":
            raise ValueError(f"output name {output_name} is missing the '.mp4' extension! ({output_name[-4:]})")
    if not output_name:
        output_name = f"COMBINED - {str(datetime.date.today())} - {v1.stem}-{v2.stem}.mp4"
    # We can change this in the config later on if we want to make this configurable
    export_path = f"{v1.folder_path}/{output_name}"

    VideoEditor.join_and_save(v1, v2, export_path)
    typer.echo(f"[red]Complete! Combined {v2.filename} to the end of {v2.filename}")


@app.command(short_help="trims a video. takes index, start (in seconds), end or duration (seconds), and outputs name.")
def trim(
        video_index: int,
        start: int,
        end: Optional[int] = None,
        duration: Optional[int] = None,
        output_name: Optional[str] = None
):
    videos = ApplicationConfig.pull_from_directory(recurse=True)
    v1 = videos[video_index]

    if output_name:
        if output_name[-4:] != ".mp4":
            raise ValueError(f"output name {output_name} is missing the '.mp4' extension!")
    if not output_name:
        output_name = f"Trimmed - {str(datetime.date.today())} - {v1.stem} - from {start}.mp4"
    # We can change this in the config later on if we want to make this configurable
    export_path = f"{v1.folder_path}/{output_name}"

    VideoEditor.cut_and_save(v1, export_path, start, end, duration)

    typer.echo(f"Complete!")
    typer.echo(f"Trimmed {v1.filename}: (start: {start}, end: {end}, duration: {duration})")
    typer.echo(f"Saved to {export_path}")


@app.command(short_help="plays the video. takes an index integer as an argument. Shuffles if left blank")
def play(video_index: int, shuffle: Optional[bool] = False):
    videos = ApplicationConfig.pull_from_directory(recurse=True)

    if shuffle:
        if len(videos) == 0:
            raise IndexError(f"There are no videos in the folder!")
        video_index = random.randint(0, len(videos) - 1)
    typer.launch(videos[video_index].filepath)

    video_text = Text(f"Opened: {videos[video_index].filename} ({video_index})", justify="center")
    video_text.stylize("bold green on white", 0, 7)
    video_text.stylize("bold red on white", 8)
    console.print(video_text, justify="center")


@app.command(
    help="shows the video list. first column is the index integer used for the app. When sorting, any space "
         "needs to be hyphenated '-'. For example, 'date-created'.")
def show(sort: Optional[list[str]] = None, ascending: Optional[bool] = True):
    videos = ApplicationConfig.as_dataframe(recurse=True)
    console.clear()

    # Table Positioning
    table = Table(title="Available Videos", box=box.HORIZONTALS, leading=True, row_styles=["steel_blue3"])
    table_centered = Align.center(table)
    table.add_column("#")
    table.add_column("filename")
    table.add_column("full path")
    table.add_column("size")
    table.add_column("date created")
    table.add_column("date modified")
    table.add_column("date accessed")
    table.add_column("duration")
    table.add_column("resolution")

    # Dataframe Configuration
    user_sort_input_mapping = {
        "#": "idx",
        "file-name": "filename",
        "full-path": "folder_path",
        "fullpath": "folder_path",
        "size": "size",
        "date-created": "date_created",
        "date-modified": "date_modified",
        "date-accessed": "date_accessed",
        "duration": "duration",
        "resolution": "resolution",
    }
    if sort:
        try:
            sort = [user_sort_input_mapping[f.lower()] for f in sort]
            videos = videos.sort_values(by=sort, ascending=ascending)
        except KeyError:
            raise KeyError(f"{sort} is not valid! Available: {videos.columns}")

    for video in videos.itertuples():
        table.add_row(
            str(video.idx),
            str(video.filename),
            str(video.folder_path),
            str(video.size),
            str(video.date_created),
            str(video.date_modified),
            str(video.date_accessed),
            str(video.duration),
            str(video.resolution)
        )
    console.log(table_centered)


@app.command(short_help="changes the directory and overwrites the JSON file")
def change_directory():
    curr_folder = ApplicationConfig.return_current_directory()

    console.print(f"[bold cyan]({curr_folder})[/bold cyan]", new_line_start=True)
    cwd = Prompt.ask(
        f"Where would you like your working directory to be?",
        default=curr_folder
    )
    if cwd != curr_folder:
        confirmed = Confirm.ask(f"Are you sure you would like to change the default directory?", default=False)
        if confirmed:
            ApplicationConfig.change_current_directory(cwd)
            console.print(f"Changed directory to [green on white]{str(cwd)}[/green on white]!")
            return
    console.print(f"[green]Directory is the same![/green]")


@app.command(short_help="generates a thumbnail contact sheet")
def thumbnail_sheet(
        video_index: int,
        output_name: Optional[str] = None,
        columns: Optional[int] = 5,
        rows: Optional[int] = 4
):
    videos = ApplicationConfig.pull_from_directory(recurse=True)
    v1 = videos[video_index]

    allowed = [".png", "jpeg", ".bmp", "tiff"]
    if output_name:
        if output_name[-4:] not in allowed:
            raise ValueError(f"output name {output_name[-1:-4]} is NOT an approved file type!")
    if not output_name:
        output_name = f"Thumbnail Sheet - {v1.stem} - {str(datetime.date.today())}.png"
    # We can change this in the config later on if we want to make this configurable
    export_path = f"{v1.folder_path}/{output_name}"
    VideoEditor.generate_thumbnail_sheet(video=v1, output_path=export_path, columns=columns, rows=rows)

    typer.echo(f"Complete!")
    typer.echo(f"Created a Thumbnail Sheet for {v1.stem}!")
    typer.echo(f"Saved to {export_path}")


if __name__ == "__main__":
    app()
